#
# Copyright (c) 2010 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#           http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from ovirtcli.shell.cmdshell import CmdShell


class ClearCmdShell(CmdShell):
    NAME = 'clear'

    def __init__(self, context, parser):
        CmdShell.__init__(self, context, parser)

    def do_clear(self, args):
        self.context.execute_string(ClearCmdShell.NAME + ' ' + args + '\n')




#
# Copyright (c) 2010 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#           http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import types

class MultiValueDict(dict):
    '''
    Dictionary that supports multi-value based keys
    '''

    def __init__(self, *args, **kwargs):
        if args and args[0] and args[0][0] and type(args[0][0]) == types.TupleType:
            for item in args[0]:
                if len(item) >= 2:
                    self.__setitem__(item[0], item[1])
                elif len(item) == 1:
                    self.__setitem__(item[0], None)
        else:
            dict.__init__(self, *args, **kwargs)

    def __setitem__(self, key, value):
        if self.has_key(key):
            if self[key] <> None and type(self[key]) != types.ListType:
                tmp = self[key]
                super(MultiValueDict, self).__setitem__(key, [])
                self[key].append(tmp)
                self[key].append(value)
            elif self[key] == None:
                super(MultiValueDict, self).__setitem__(key, value)
            else:
                self[key].append(value)
        else:
            super(MultiValueDict, self).__setitem__(key, value)



#
# Copyright (c) 2013 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#           http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import types


class Requires(object):
    """
    Checks that method arg is of a given type

    @note:

    1. checks that method's argument of type DFSAEvent

        @Requires(DFSAEvent)
        def method1(self, event):
            ...

    2. checks that method's argument is a List of DFSAEvent

        @Requires([DFSAEvent])
        def method2(self, events):
            ...
    """

    def __init__(self, types_to_check):
        """
        Checks that method arg is of a given type

        @param types_to_check: the types to validate against
        @note:

        1. checks that method's argument of type DFSAEvent

            @Requires(DFSAEvent)
            def method1(self, event):
                ...

        2. checks that method's argument is a List of DFSAEvent

            @Requires([DFSAEvent])
            def method2(self, events):
                ...
        """
        assert types_to_check != None
        self.__types_to_check = types_to_check

    def __call__(self, original_func):
        decorator_self = self
        def wrappee(*args, **kwargs):
            self.__check_list(
                      args[1:],
                      decorator_self.__types_to_check
            )
            return original_func(*args, **kwargs)
        return wrappee

    def __raise_error(self, typ):
        raise TypeError(
                "%s instance is expected."
                %
                typ.__name__
        )

    def __check_list(self, candidates, typs):
        if isinstance(typs, types.ListType):
            if type(candidates[0]) is types.ListType:
                # the list of items of a specific type
                if candidates[0]:
                    for candidate in candidates[0]:
                        if not isinstance(candidate, typs[0]):
                            self.__raise_error(typs[0])
                else:
                    self.__raise_error(types.ListType)
        else:
            # the items is of a specific type
            if not isinstance(candidates[0], typs):
                self.__raise_error(typs)



#
# Copyright (c) 2010 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#           http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from kitchen.text.converters import getwriter


class Terminal(object):
    """Base class for terminal objects."""

    width = None
    height = None

    def __init__(self, stdin, stdout, stderr):
        self.stdin = stdin
        self.stdout = getwriter('utf8')(stdout)
        self.stderr = stderr

    def clear(self):
        raise NotImplementedError

    def set_echo(self, echo):
        raise NotImplementedError

    def readline(self, prompt):
        raise NotImplementedError




import sys

if sys.platform in ('linux2',):
    from src.ovirtcli.platform.posix import util
    from src.ovirtcli.platform.posix import vnc
    from src.ovirtcli.platform.posix import spice

elif sys.platform in ('win32',):
    pass



#
# Copyright (c) 2010 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#           http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import code
import readline
import atexit
import os
from ovirtcli.infrastructure.settings import OvirtCliSettings
import threading

HISTORY_METAFILE = "~/." + OvirtCliSettings.PRODUCT.lower() + "shellhistory"
TMP_HISTORY_METAFILE = HISTORY_METAFILE + ".tmp"

class HistoryManager(code.InteractiveConsole):
    """ Provides history management capabilities """

    def __init__(self,
                 locals=None,  # @ReservedAssignment
                 filename="<console>",
                 histfile=os.path.expanduser(HISTORY_METAFILE),
                 temp_histfile=os.path.expanduser(TMP_HISTORY_METAFILE),
                 enabled=False):

        readline.set_history_length(3000)
        self.lock = threading.RLock()
        self.enabled = enabled
        self.histfile = histfile
        self.tmp_histfile = temp_histfile
        code.InteractiveConsole.__init__(self, locals, filename)
        if enabled: self.enable()

    def get(self, indx):
        return readline.get_history_item(indx)

    def length(self):
        return readline.get_current_history_length()

    def list(self):
        buff = []
        ln = self.length()
        if ln > 0:
            for i in range(ln):
                if i > 0: buff.append(self.get(i))
        return buff

    def listBetweenIndexes(self, from_indx, to_indx):
        buff = []
        ln = self.length()
        if ln > 0:
            for i in range(from_indx, to_indx):
                if i > 0 and i < ln: buff.append(self.get(i))
        return buff

    def export(self, filename):
        readline.write_history_file(filename)

    def clear(self):
        readline.clear_history()

    def enable(self):
        with self.lock:
            self.enabled = True
            self.__register_file(self.histfile)

    def disable(self):
        with self.lock:
            self.enabled = False
            self.__dump_callback(self.histfile)
            self.__register_file(self.tmp_histfile)
            self.clear()

    def remove(self, entry):
        if self.length() > 0:
            readline.remove_history_item(entry)

    def __unregister_dump_callback(self):
        for item in atexit._exithandlers:
            if hasattr(item[0], 'func_name') and \
               item[0].func_name == '__dump_callback':
                atexit._exithandlers.remove(item)

    def __register_file(self, filename):
        self.__unregister_dump_callback()
        if hasattr(readline, "read_history_file"):
            try:
                readline.read_history_file(filename)
            except IOError:
                pass
            atexit.register(self.__dump_callback, filename)

    def __dump_callback(self, histfile):
        readline.write_history_file(histfile)




#
# Copyright (c) 2010 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#           http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from cli.settings import Settings, enum, boolean


class OvirtCliSettings(Settings):

    PRODUCT = 'oVirt'

    INTRO = \
    """Welcome to %s shell""" % PRODUCT

    DISCONNECTED_TEMPLATE = \
""">>> disconnected from %s manager <<<""" % PRODUCT

    CONNECTED_TEMPLATE = \
">>> connected to " + PRODUCT + " manager %s <<<"

    settings = Settings.settings + [
        ('ovirt-shell:name', str, '%s-shell' % PRODUCT),
        ('ovirt-shell:url', str, ''),
        ('ovirt-shell:username', str, ''),
        ('ovirt-shell:password', str, ''),
        ('ovirt-shell:key_file', str, None),
        ('ovirt-shell:cert_file', str, None),
        ('ovirt-shell:ca_file', str, None),
        ('ovirt-shell:insecure', boolean, False),
        ('ovirt-shell:dont_validate_cert_chain', boolean, False),
        ('ovirt-shell:filter', boolean, False),
        ('ovirt-shell:port', int, None),
        ('ovirt-shell:timeout', int, None),
        ('ovirt-shell:session_timeout', int, None),
        ('ovirt-shell:input_format', enum('xml'), 'xml'),
        ('ovirt-shell:output_format', enum('xml', 'text'), 'text'),
        ('ovirt-shell:wide', boolean, False),
        ('ovirt-shell:header', boolean, True),
        ('ovirt-shell:fields', str, None),
        ('ovirt-shell:fields.*', str, None),
        ('ovirt-shell:ps1.connected', str, '[' + PRODUCT + ' shell %(version)s (connected)]# '),
        ('ovirt-shell:ps2.connected', str, '[' + PRODUCT + ' shell (connected)]# '),
        ('ovirt-shell:ps3.connected', str, '[' + PRODUCT + ' shell (connected@%(host)s)]# '),
        ('ovirt-shell:ps1.disconnected', str, '[%s shell (disconnected)]# ' % PRODUCT),
        ('ovirt-shell:ps1.unauthorized', str, '[%s shell (unauthorized)]# ' % PRODUCT),
        ('ovirt-shell:commands', str, '%s shell commands:' % PRODUCT),
        ('ovirt-shell:misc_commands', str, '%s shell commands:' % PRODUCT),
        ('ovirt-shell:version', str, ''),
        ('ovirt-shell:prompt', str, ''),
        ('ovirt-shell:extended_prompt', boolean, False),
        ('ovirt-shell:execute_command', str, None),
        ('ovirt-shell:renew_session', boolean, False)
    ]

    # config file white list
    config_items = [
        'cli:autopage',
        'cli:autoconnect',
        'ovirt-shell:url',
        'ovirt-shell:username',
        'ovirt-shell:password',
        'ovirt-shell:key_file',
        'ovirt-shell:cert_file',
        'ovirt-shell:ca_file',
        'ovirt-shell:insecure',
        'ovirt-shell:dont_validate_cert_chain',
        'ovirt-shell:filter',
        'ovirt-shell:timeout',
        'ovirt-shell:session_timeout',
        'ovirt-shell:extended_prompt',
        'ovirt-shell:renew_session'
    ]



#
# Copyright (c) 2013 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#           http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class Singleton(type):
    """
    The singleton annotation

    applying this annotation on class, will make sure that no
    more than single instance of this class can be created.
    """
    __all_instances = {}
    def __call__(cls, *args, **kwargs):  # @NoSelf
        if cls not in cls.__all_instances:
            cls.__all_instances[cls] = super(
                         Singleton,
                         cls).__call__(
                               *args,
                               **kwargs
                         )
        return cls.__all_instances[cls]



#
# Copyright (c) 2010 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#           http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from ovirtcli.shell.cmdshell import CmdShell
from ovirtcli.utils.typehelper import TypeHelper
from ovirtcli.utils.autocompletionhelper import AutoCompletionHelper
from ovirtsdk.infrastructure import brokers
from ovirtcli.utils.methodhelper import MethodHelper


class ShowCmdShell(CmdShell):
    NAME = 'show'
    ALIAS = 'get'

    def __init__(self, context, parser):
        CmdShell.__init__(self, context, parser)

    def do_show(self, args):
        return self.context.execute_string(ShowCmdShell.NAME + ' ' + args + '\n')

    def __add_resource_specific_options(self, obj, specific_options, line, key=None):
        obj_coll_type = TypeHelper.getDecoratorType(TypeHelper.to_plural(obj))
        if obj_coll_type:
            if hasattr(brokers, obj_coll_type):
                obj_coll = getattr(brokers, obj_coll_type)
                if obj_coll and hasattr(obj_coll, ShowCmdShell.ALIAS):
                    method_args = MethodHelper.get_documented_arguments(method_ref=getattr(obj_coll, ShowCmdShell.ALIAS),
                                                                        as_params_collection=True,
                                                                        spilt_or=True)
                    if method_args:
                        specific_options[obj if key == None else key] = method_args

    def complete_show(self, text, line, begidx, endidx):

        args = TypeHelper.get_types_by_method(False, ShowCmdShell.ALIAS, expendNestedTypes=True)
        specific_options = self.get_resource_specific_options(args,
                                                              line,
                                                              callback=self.__add_resource_specific_options)
        return AutoCompletionHelper.complete(line=line,
                                             text=text,
                                             args=args,
                                             specific_options=specific_options)
    def is_show_argument(self, line, key):
        args = TypeHelper.get_types_by_method(False, ShowCmdShell.ALIAS, expendNestedTypes=True)
        if key in args:
            return True
        specific_options = self.get_resource_specific_options(args,
                                                              line,
                                                              callback=self.__add_resource_specific_options)
        for arg_key in specific_options.keys():
             if key in specific_options[arg_key]:
                 return True
        return False




#
# Copyright (c) 2010 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#           http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import os
import re
import stat

from fnmatch import fnmatch
from ConfigParser import ConfigParser
from cli import platform
import types
import sys


class enum(object):
    """A setting that can have one of a predetermined set of values."""

    def __init__(self, *values):
        self.values = values

    def __call__(self, value):
        if value not in self.values:
            raise ValueError, 'illegal value: %s' % value
        return value


class regex(object):
    """A setting that is matched against a regular expression."""

    def __init__(self, regex):
        self.regex = re.compile(regex)

    def __call__(self, value):
        if not self.regex.match(value):
            raise ValueError, 'illegal value: %s' % value
        return value


def boolean(value):
    """A boolean setting."""
    if isinstance(value, int) or isinstance(value, bool):
        return bool(value)
    elif isinstance(value, str):
        s = value.lower()
        if s in ('true', 'on', '1'):
            return True
        elif s in ('false', 'off', '0'):
            return False
    raise ValueError, 'illegal value: %s' % value


class Settings(dict):
    """Base class for settings."""

    settings = [
        ('cli:ps1', str, '$ '),
        ('cli:ps2', str, '> '),
        ('cli:debug', boolean, False),
        ('cli:verbosity', int, 0),
        ('cli:autopage', boolean, True),
        ('cli:autoconnect', boolean, True),
        ('cli:pager', str, None)
    ]

    def __init__(self, name):
        """Constructor."""
        self.name = name
        self.callbacks = []
        self.update(self.get_defaults())

    def __setitem__(self, key, value):
        """Validate a variable. Also calls callbacks."""
        found = False
        for pattern, validator, default in self.settings:
            if not fnmatch(key, pattern):
                continue
            value = self.__normalazie_value(value)
            if value:
                try:
                    value = validator(value)
                except ValueError:
                    # delegate type related errors handling to SDK
                    pass
            found = True
        if not found:
            raise KeyError, 'unknown setting: %s' % key
        for pattern, callback in self.callbacks:
            if not fnmatch(key, pattern):
                continue
            callback(key, value)
        super(Settings, self).__setitem__(key, value)

    def __normalazie_value(self, value):
        """Converts string value to python type """
        if value:
            if type(value) == types.StringType and value == 'None':
                return None
        return value

    def get_defaults(self):
        """Return a dictionary with the default settings."""
        return dict(((p, d) for p, t, d in self.settings
                     if '*' not in p))

    def load_config_file(self):
        """
        Load default values from a configuration file.

        @return: if-file-exist, is-config-file-in-old-format
        """
        old_format = False
        fname = platform.local_settings_file(self.name)
        from ovirtcli.infrastructure.settings import OvirtCliSettings
        if fname is None:
            return False, old_format
        cp = ConfigParser()
        if not cp.read(fname):
            return False, old_format
        for section in cp.sections():
            items = cp.items(section)
            if len(items) != len(OvirtCliSettings.config_items):
                old_format = True
            for key, value in items:
                conf_key = '%s:%s' % (section, key)
                if conf_key not in OvirtCliSettings.config_items:
                    old_format = True
                else:
                    self[conf_key] = value

        return True, old_format

    def _write_config_file(self, settings, example=False):
        """Overwrite the configuration file with the current settings."""
        fname = platform.local_settings_file(self.name)
        if fname is None:
            return
        ftmp = '%s.%d-tmp' % (fname, os.getpid())
        fout = file(ftmp, 'w')
        sections = {}
        from ovirtcli.infrastructure.settings import OvirtCliSettings
        for key in settings:
            if key in OvirtCliSettings.config_items:
                section, name = key.split(':')
                if section not in sections:
                    sections[section] = {}
                sections[section][name] = settings[key]
        for section in sorted(sections):
            fout.write('[%s]\n' % section)
            for key in sections[section]:
                if example:
                    fout.write('#')
                fout.write('%s = %s\n' % (key, sections[section][key]))
        fout.close()
        self.set_file_permissions(ftmp)
        os.rename(ftmp, fname)

    def set_file_permissions(self, f):
        # Set UID bit
        # Owner has read permission
        # Owner has write permission
        # Do not dump the file.
        os.chmod(f,
                 stat.S_ISGID |
                 stat.S_IRUSR |
                 stat.S_IWUSR |
                 stat.UF_NODUMP)

    def write_config_file(self):
        """Overwrite the config file with the current settings."""
        self._write_config_file(self, False)

    def write_example_config_file(self):
        """Write an example config file."""
        self._write_config_file(self.get_defaults(), True)

    def add_callback(self, pattern, callback):
        """Register a callback function. The callback is called when the
        variable identified by `pattern' changes."""
        self.callbacks.append((pattern, callback))



#
# Copyright (c) 2010 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#           http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import re

from ovirtcli.command.command import OvirtCommand
from ovirtsdk.api import API
from ovirtsdk.infrastructure.errors import RequestError, NoCertificatesError, \
    ConnectionError
from cli.messages import Messages
from urlparse import urlparse
from ovirtcli.shell.connectcmdshell import ConnectCmdShell
from ovirtcli.state.statemachine import StateMachine

class ConnectCommand(OvirtCommand):

    name = 'connect'
    description = 'connect to a oVirt manager'
    args_check = (0, len(ConnectCmdShell.OPTIONS))
    valid_options = [ (
          '--' + item, str
         )
         for item in ConnectCmdShell.OPTIONS
    ]

    helptext = """\
        == Usage ==

        connect [command options]

        == Description ==

        Connect to a oVirt manager. This command has two forms. In the first
        form, no arguments are provided, and the connection details are read
        from their respective configuration variables. In the second form,
        the connection details are provided as arguments.

        == Arguments ==

         * url               - The URL to connect to (http[s]://server[:port]/ovirt-engine/api).
         * username          - The user to connect as. (user@domain).
         * password          - The password to use.
         * [key-file]        - The client PEM key file to use.
         * [cert-file]       - The client PEM certificate file to use.
         * [ca-file]         - The server CA certificate file to use.
         * [filter]          - Enables user permission based filtering.
         * [insecure]        - Allow connecting to SSL sites without certificates.
         * [port]            - The port to use (if not specified in url).
         * [timeout]         - The request timeout.
         * [session-timeout] - The authentication session timeout (positive number).
        """

    def execute(self):
        args = self.arguments
        settings = self.context.settings
        context = self.context

        MIN_FORCE_CREDENTIALS_CHECK_VERSION = ('00000003', '00000001', '00000000', '00000004')

        key_file = self.xNoneType(settings.get('ovirt-shell:key_file'))
        cert_file = self.xNoneType(settings.get('ovirt-shell:cert_file'))
        ca_file = self.xNoneType(settings.get('ovirt-shell:ca_file'))
        port = settings.get('ovirt-shell:port')
        timeout = settings.get('ovirt-shell:timeout')
        session_timeout = settings.get('ovirt-shell:session_timeout')
        renew_session = settings.get('ovirt-shell:renew_session')
        debug = settings.get('cli:debug')
        insecure = settings.get('ovirt-shell:insecure')
        dont_validate_cert_chain = settings.get('ovirt-shell:dont_validate_cert_chain')
        filter_ = settings.get('ovirt-shell:filter')

        if self.context.connection is not None and \
           self.context.status != self.context.COMMUNICATION_ERROR and \
           self.context.status != self.context.AUTHENTICATION_ERROR and \
           self.__test_connectivity():
            self.write(
                   Messages.Warning.ALREADY_CONNECTED
            )
            return
        if len(args) == 3:
            url, username, password = args
        else:
            url = settings.get('ovirt-shell:url')
            if not url:
                self.error(
                       Messages.Error.MISSING_CONFIGURATION_VARIABLE % 'url'
                )
            username = settings.get('ovirt-shell:username')
            if not username:
                self.error(
                       Messages.Error.MISSING_CONFIGURATION_VARIABLE % 'username'
                )
            password = settings.get('ovirt-shell:password')
            if not password:
                self.error(
                   Messages.Error.MISSING_CONFIGURATION_VARIABLE % 'password'
                )

        if not self.is_valid_url(url):
            self.error(
               Messages.Error.INVALID_URL_SEGMENT % url
            )

        try:
            StateMachine.connecting()  # @UndefinedVariable

            self.context.set_connection (
                     API(
                         url=url,
                         username=username,
                         password=password,
                         key_file=key_file,
                         cert_file=cert_file,
                         ca_file=ca_file,
                         insecure=insecure,
                         validate_cert_chain=not dont_validate_cert_chain,
                         filter=filter_,
                         port=port if port != -1 else None,
                         timeout=timeout if timeout != -1 else None,
                         session_timeout=session_timeout if session_timeout != -1 else None,
                         renew_session=renew_session,
                         debug=debug
                     ),
                     url=url
             )

            if context.sdk_version < MIN_FORCE_CREDENTIALS_CHECK_VERSION:
                self.__test_connectivity()

            StateMachine.connected()  # @UndefinedVariable

        except RequestError, e:
            StateMachine.rollback()
            self.__cleanContext()
            if debug:
                self.error("[" + str(e.status) + '] - ' + str(e.reason) + ', ' + str(e.detail))
            else:
                self.error("[" + str(e.status) + '] - ' + str(e.reason))
        except NoCertificatesError:
            StateMachine.rollback()
            self.__cleanContext()
            self.error(Messages.Error.NO_CERTIFICATES)
        except ConnectionError, e:
            StateMachine.rollback()
            self.__cleanContext()
            self.context._clean_settings()
            self.error(str(e))
        except TypeError, e:
            StateMachine.rollback()
            self.__cleanContext()
            option, value, expected = self.__normalize_typeerror(e)
            self.error(
                   Messages.Error.INVALID_ARGUMENT_TYPE % (
                               option,
                               value,
                               expected)
            )
        except Exception, e:
            StateMachine.rollback()
            self.__cleanContext()
            self.error(str(e))
        finally:
            # do not log connect command details as it may be
            # a subject for password stealing or DOS attack
            self.__remove_history_entry()

    def is_valid_url(self, url):
        if url is None:
            return False;
        regex = re.compile(
            r'^(?:http)s?://'  # http:// or https://
            r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
            r'localhost|'  # localhost...
            r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}|'  # ...or ipv4
            r'\[?[A-F0-9]*:[A-F0-9:]+\]?)'  # ...or ipv6
            r'(?::\d+)?'  # optional port
           , re.IGNORECASE)
        if not regex.search(url):
            return False
        hostname = self.get_hostname(url)
        if not self.hostname_is_ip(hostname):
            return True
        return self.is_valid_ip(hostname)

    def get_hostname(self, url):
        url_obj = urlparse(url)
        return getattr(url_obj, 'hostname')

    def hostname_is_ip(self, hostname):
        regex = re.compile(
                   r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}',
                   re.IGNORECASE
        )
        return regex.search(hostname)

    def is_valid_ip(self, hostip):
        regex = re.compile(
                   r'^((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$',
                   re.IGNORECASE
        )
        return regex.search(hostip)

    def __normalize_typeerror(self, exception):
            err = str(exception)
            error = err[1:len(err) - 1]
            error = error.replace("\"", "")
            error = error.replace("'", "")
            splitted = error.split(', ')
            option_value = splitted[0].split('=')
            option = option_value[0].upper()
            value = option_value[1]
            expected = splitted[1]

            return option, value, expected

    def __test_connectivity(self):
        return self.context.connection.test(throw_exception=True)

    def __remove_history_entry(self):
        last_entry = self.context.history.get(self.context.history.length() - 1)
        if last_entry and last_entry.startswith(ConnectCommand.name):
            self.context.history.remove(self.context.history.length() - 1)

    def __cleanContext(self):
        if self.context.connection is not None:
            self.context._clean_settings()
            try:
                self.context.connection.disconnect()
            except Exception, e:
                self.error(e.strerror.lower())
            finally:
                self.context.connection = None

    def xNoneType(self, s):
        return None if s == 'None' else s



#
# Copyright (c) 2010 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#           http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import shlex
import sys

from ovirtcli.shell.cmdshell import CmdShell
from ovirtcli.utils.autocompletionhelper import AutoCompletionHelper

class ConnectCmdShell(CmdShell):
    NAME = 'connect'
    OPTIONS = [
       'url',
       'user',
       'password',
       'key-file',
       'cert-file',
       'ca-file',
       'insecure',
       'filter',
       'port',
       'timeout',
       'session-timeout'
    ]

    def __init__(self, context, parser):
        CmdShell.__init__(self, context, parser)

    def __do_verbose_connect(self, context, parser, opts):
        if not self.copy_environment_vars(context):
            sys.exit(1)
        if not self.copy_cmdline_options(opts, context, parser):
            sys.exit(1)
        self.context.execute_string(ConnectCmdShell.NAME + '\n')

    def __do_connect(self, args):
        arg = ConnectCmdShell.NAME + ' ' + args.strip() + '\n'
        self.context.execute_string(arg)

    def do_connect(self, args):
        arg = '--connect ' + args.strip() + '\n'
        m_opts, m_args = self.parser.parse_args(args=shlex.split(arg))

        if m_opts.connect and len(m_args) == 0:
            self.__do_verbose_connect(self.context, self.parser, m_opts)
        else:
            self.__do_connect(args)

    def complete_connect(self, text, line, begidx, endidx):
        return AutoCompletionHelper.complete(line=line, text=text,
                                             args={}.fromkeys(ConnectCmdShell.OPTIONS),
                                             all_options=True)

    def is_connect_argument(self, line, key):
        return key in ConnectCmdShell.OPTIONS



#
# Copyright (c) 2010 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#           http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import os.path
import sys
import pwd

stdin_fileno = sys.stdin.fileno()
stdout_fileno = sys.stdout.fileno()
stderr_fileno = sys.stderr.fileno()


def which(cmd):
    """Find a command `cmd' in the path."""
    if cmd.startswith('/') and os.access(cmd, os.X_OK):
        return cmd
    path = os.environ.get('PATH')
    path = path.split(os.pathsep)
    for dir in path:
        fname = os.path.join(dir, cmd)
        if os.access(fname, os.X_OK):
            return fname


def spawn(cmd, args, debug=False):
    """Spawn a command. Return a (pid, stdin). If debug is False, stdout and
    stderr will be redirected to /dev/null."""
    stdin, pstdin = os.pipe()
    pid = os.fork()
    if pid > 0:
        os.close(stdin)
        return (pid, pstdin)
    # Put ourselves in a new session so that we don't get any signal related
    # to the terminal that ovirt-shell is running in.
    os.setsid()
    os.close(pstdin)
    if stdin != stdin_fileno:
        os.close(stdin_fileno)
        os.dup2(stdin, stdin_fileno)
    if not debug:
        os.close(stdout_fileno)
        stdout = os.open('/dev/null', os.O_WRONLY)
        if stdout != stdout_fileno:
            os.dup2(stdout, stdout_fileno)
            os.close(stdout)
        os.close(stderr_fileno)
        stderr = os.open('/dev/null', os.O_WRONLY)
        if stderr != stderr_fileno:
            os.dup2(stderr, stderr_fileno)
            os.close(stderr)
    os.execv(cmd, args)
    # exec failed.. the error message only gets through if debug=True
    os.write(stderr_fileno, 'execv() failed')
    sys.exit(1)


def get_home_dir():
    """Return the user's home directory."""
    home = os.environ.get('HOME')
    if home is not None:
        return home
    try:
        pw = pwd.getpwuid(os.getuid())
    except Exception:
        return None
    return pw.pw_dir



#
# Copyright (c) 2010 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#           http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class Formatter(object):
    """Base class for formatter objects."""

    name = None

    def format(self, context, result, scope=None):
        raise NotImplementedError

    def format_terminal(self, text, border, termwidth, newline="\n\n", header=None, offsettext=True):
        """
        formats (pretty) screen width adapted messages with border
        
        @param text: text to prin
        @param border: border to use
        @param termwidth: terminal width
        @param newline: new line separator (default is '\n\n')
        @param header: upper border header (default is None)
        @param offsettext: align the text to middle of the screen (default True)
        """

        linlebreak = '\r\n'
        offset = "  "
        space = " "
        introoffset = (termwidth / 2 - (len(text) / 2))
        borderoffset = (termwidth - 4)

        # align multilined output
        if text.find(linlebreak) <> -1 :
            offsettext = False
            text = offset + text.replace(linlebreak, (linlebreak + offset))

        if (header):
            headeroffset = (borderoffset / 2 - ((len(header) / 2)))
            oddoffset = 0 if termwidth & 1 != 0 else 1
            return offset + headeroffset * border + space + header + space + \
                   (headeroffset - len(offset) - oddoffset) * border + newline + \
                   ((introoffset * space)  if offsettext else "") + text + newline + \
                   offset + borderoffset * border + newline
        return offset + borderoffset * border + newline + \
               ((introoffset * space) if offsettext else "") + text + newline + \
               offset + borderoffset * border + newline




#
# Copyright (c) 2010 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#           http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from ovirtcli.command.command import OvirtCommand
from cli.messages import Messages


class StatusCommand(OvirtCommand):

    name = 'status'
    description = 'shows last command status'
    helptext = """\
        == Usage ==

        status

        == Description ==

        Shows the exist status of the last command.
        """

    def execute(self):
        context = self.context
        status = context.status
        if status is not None:
            sstatus = str(status)
            for sym in dir(context):
                if sym[0].isupper() and getattr(context, sym) == status:
                    sstatus += ' (%s)' % sym
        else:
            sstatus = 'N/A'
        self.write(Messages.Info.STATUS % sstatus + '\n')




#
# Copyright (c) 2010 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#           http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from ovirtcli.command.command import OvirtCommand
from ovirtcli.utils.typehelper import TypeHelper

from ovirtsdk.utils.parsehelper import ParseHelper
import types
from cli.messages import Messages

class ActionCommand(OvirtCommand):

    name = 'action'
    description = 'execute an action on an object'
    usage = 'action <type> <id> <action> [parent identifiers] [command options]'
    args_check = 3
    valid_options = [ ('*', str) ]

    helptext0 = """\
        == Usage ==

        action <type> <id> <action> [parent identifiers] [command options]

        Note that when using auto-completion it may be necessary to know
        the parent identifiers of the object in order to determine the
        available actions, so the following syntax, with the action
        after the parent identifiers, is also accepted:

        action <type> <id> [parent identifiers] <action> [command options]

        == Description ==

        Executes the an action on an object. This command requires the
        following arguments:

          * type        - The type to operate on
          * id          - The name or id identifying the object
          * action      - The action to take

        For more specific help on the available actions and options, use
        'help action <type> <id>'

        == Supported Help formats ==

        - This help will list all available attribute options for given resource
          
          * format      - help action type id
          * example     - help action vm myvm

        - This help will list all available attribute options for specific action on given resource
        
          * format      - help action type id action_name
          * example     - help action vm myvm start
        
        - This help will list all available attribute options for specific subresource 
        
          * format      - help action resource resource_name --subresource-identifier mysubresource
          * example     - help action nic bond0 --host-identifier myhost
        
        - This help will display all available attribute options for specific action on given subresource

          * format      - help action action_name resource resource_name --subresource-identifier mysubresource
          * example     - help action attach nic bond0 --host-identifier myhost

        == Available types ==

        The <type> parameter must be one of the following:

          $types

        == Return values ==

        This command will return one of the following statuses. To see the
        exit status of the last command, type 'status'.

          $statuses
        """

    helptext1 = """\
        == Usage ==

        action <type> <id> <action> [parent identifiers] [command options]

        Note that when using auto-completion it may be necessary to know
        the parent identifiers of the object in order to determine the
        available actions, so the following syntax, with the action
        after the parent identifiers, is also accepted:

        action <type> <id> [parent identifiers] <action> [command options]

        == Description ==

        Executes the an action on an object. This command requires the
        following arguments:

          * type        - The type to operate on
          * id          - The name or id identifying the object
          * action      - The action to take

        == Available actions ==

        The following actions are available for the specified $type object:

          $actions

        == Collection based option format ==

          * [--x-y: collection]
            {
              [y.a: string]
              [y.b: string]
              [y.c: string]
            }

          e.g:

          --x-y "y.a=a1,y.b=b1,y.c=c1"
          --x-y "y.a=a2,y.b=b2,y.c=c2"
          ...

          where a, b, c are option properties and aN, bN, cN is actual user's data

        == Object Identifiers ==

        Some objects can only exist inside other objects. For example, a disk
        can only exist in the content of a virtual machine. In this case, one
        or more object identifiers needs to be provided to identify the
        containing object.

        An object identifier is an option of the form '--<type>id <id>'. This
        would identify an object with type <type> and id <id>. See the
        examples section below for a few examples.

        == Attribute Options ==

        The following attribute options are understood. Note: this lists all
        available attribute options for action '$action'.

          $options

        == Examples ==

        - This example starts a vm named "myvm" with few optional parameters:

          $ action vm iscsi_desktop start --vm-display-type vnc --vm-cpu-topology-sockets 2 --vm-cpu-topology-cores 2

        - This example detaches a host nic with id 'mynic' from host 'myhost':

          $ action nic mynic detach --host-identifier myhost
    
        == Return values ==

        This command will exit with one of the following statuses. To see the
        exit status of the last command, type 'status'.

          $statuses
        """

    def execute(self):
        """Execute the action command."""
        args = self.arguments
        opts = self.options

        if not (TypeHelper.isKnownType(args[0])) and \
           ParseHelper.getXmlType(args[0]) == None:
            self.error(
                       Messages.Error.NO_SUCH_TYPE % args[0]
            )

        # scope = '%s:%s' % (ParseHelper.getXmlWrapperType(args[0]), args[2])
        actionable_types = TypeHelper.get_actionable_types(
                       expendNestedTypes=True,
                       groupOptions=True
        )

        resource = self.get_object(
                       args[0],
                       args[1],
                       self.resolve_base(opts),
                       context_variants=actionable_types[args[0]]
        )

        if resource is None:
            self.error(
                   Messages.Error.NO_SUCH_OBJECT % (args[0], args[1])
            )
        elif hasattr(resource, args[2]) and \
             type(getattr(resource, args[2])) == types.MethodType:
            try:
                result = self.execute_method(resource, args[2], opts)
            except Exception, e:
                self.error(str(e))
            if result.status.state == 'failed':
                self.error(
                       Messages.Info.ACTION_STATUS % result.status.state
                )
        else:
            self.error(
                   Messages.Error.NO_SUCH_ACTION % args[2]
            )
        self.context.formatter.format(self.context, result)

    def show_help(self):
        """Show help for the action command."""
        args = self.arguments
        opts = self.options
        types = TypeHelper.get_actionable_types(
                                expendNestedTypes=True,
                                groupOptions=True
        )
        subst = {}

        if not args or self.is_supported_type(types.keys(), args[0]):
            if len(args) == 2 and len(opts) == 2:
                helptext = self.helptext1

                subst['type'] = args[0]
                subst['id'] = args[1]

                base = self.resolve_base(self.options)
                obj = self.get_object(
                              args[0],
                              args[1],
                              base,
                              context_variants=types[args[0]]
                )
                if obj is None:
                    self.error(
                           Messages.Error.NO_SUCH_OBJECT % (args[1], args[1])
                    )
                actions = self._get_action_methods(obj)
                subst['actions'] = self.format_list(actions)

            if len(args) == 3 and len(opts) == 2:
                helptext = self.helptext1

                subst['type'] = args[0]
                subst['id'] = args[1]
                subst['action'] = args[0]

                base = self.resolve_base(self.options)
                obj = self.get_object(
                              args[1],
                              args[2],
                              base,
                              context_variants=types[args[0]]
                )
                if obj is None:
                    self.error(
                       Messages.Error.NO_SUCH_OBJECT % (args[0], args[1])
                    )

                actions = self._get_action_methods(obj)
                if args[0] not in actions:
                    self.error(
                           Messages.Error.NO_SUCH_ACTION % args[2]
                    )

                options = self.get_options(
                           method=args[0],
                           resource=obj,
                           context_variants=types[args[0]]
                )
                subst['actions'] = self.format_list(actions)
                subst['options'] = self.format_list(options, bullet='', sort=False)

            elif len(args) == 1:
                helptext = self.helptext0
                subst['types'] = self.format_map({args[0]:types[args[0]]})
                subst['type'] = args[0]

            elif len(args) == 2:
                helptext = self.helptext1

                subst['type'] = args[0]
                subst['id'] = args[1]
                base = self.resolve_base(opts)
                obj = self.get_object(
                              args[0],
                              args[1],
                              base,
                              context_variants=types[args[0]]
                )
                if obj is None:
                    self.error(
                       Messages.Error.NO_SUCH_OBJECT % (args[0], args[1])
                )
                actions = self._get_action_methods(obj)
                subst['actions'] = self.format_list(actions)

            elif len(args) == 3:
                helptext = self.helptext1

                subst['type'] = args[0]
                subst['id'] = args[1]
                subst['action'] = args[2]

                base = self.resolve_base(self.options)
                obj = self.get_object(
                          args[0],
                          args[1],
                          base,
                          context_variants=types[args[0]]
                )
                if obj is None:
                    self.error(
                       Messages.Error.NO_SUCH_OBJECT % (args[0], args[1])
                    )

                actions = self._get_action_methods(obj)
                if args[2] not in actions:
                    self.error(Messages.Error.NO_SUCH_ACTION % args[2])

                options = self.get_options(
                           method=args[2],
                           resource=obj,
                           sub_resource=base,
                           context_variants=types[args[0]]
                )
                subst['actions'] = self.format_list(actions)
                subst['options'] = self.format_list(options, bullet='', sort=False)

            else:
                helptext = self.helptext0
                subst['types'] = self.format_map(types)

    #            scope = '%s:%s' % (type(obj).__name__, args[2])

            statuses = self.get_statuses()
            subst['statuses'] = self.format_list(statuses)
            helptext = self.format_help(helptext, subst)
            self.write(helptext)



#
# Copyright (c) 2010 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#           http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import sys

def create(cls, *args, **kwargs):
    """This is our default constructor. It takes care of instantiating the
    right subclass, and configuring it in the right way."""
    from cli.terminal import Terminal
    from optparse import OptionParser
    if issubclass(cls, Terminal):
        obj = cls(sys.stdin, sys.stdout, sys.stderr, **kwargs)
    elif issubclass(cls, OptionParser):
        obj = cls(*args, **kwargs)
        obj.disable_interspersed_args()
    else:
        obj = cls(*args, **kwargs)
    return obj




import sys

if sys.platform in ('linux2',):
    from cli.platform.posix.terminal import PosixTerminal as Terminal
    from cli.platform.posix.util import *

elif sys.platform in ('win32',):
    pass



#
# Copyright (c) 2010 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#           http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import os
from cli.command.command import Command


class PwdCommand(Command):

    name = 'pwd'
    description = 'print working directory'
    helptext = """\
        == Usage ==

        pwd

        == Description ==
        
        Print the current working directory.
        """

    def execute(self):
        stdout = self.context.terminal.stdout
        stdout.write('%s\n' % os.getcwd())



#
# Copyright (c) 2010 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#           http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


def create_password_login_command(user, host):
    """
    Create an ssh command line.
    """
    cmd = 'ssh'
    cmd += ' -oPreferredAuthentications=password'
    cmd += ' -oStrictHostKeyChecking=no'
    cmd += ' %s@%s' % (user, host)
    return cmd



# Extract a private key from a Java key store. The standard program
# "keytool" that is part of the JRE unfortunately doesn't have this
# functionality (presumably because of 'security' reasons).
#
# The simplest way to do this, would be to convert the key store to PKCS12
# format, and then use openssl to extract the private key (OpenSSL does have
# this functionality). However for some reason, the required command
# "keytool -importkeystore -deststoretype PKCS12" often hangs for me. I was
# not able to determine the cause of this behavior.
#
# So the option taken here is to use a small Java program that uses the Java
# API to extract the private key. This means that we're adding a dependency on
# the JDK.
#
# Why this "Java in shell" hack instead of a proper Java program? Because I
# know that the system from which i created this script has the JDK installed,
# and it saves me from having to distribute compiled versions.
#
# I took inspiration for the Java program from:
# http://mark.foster.cc/pub/java/ExportPriv.java
#
# This file is licensed under the MIT license, see
# http://www.opensource.org/licenses/mit-license.php

dumpkey_function = """
function dumpkey()
{
    if [ "$#" -ne "3" ]; then
        echo "usage: $0 <keystore> <alias> <password>"
        exit 1
    fi

    tmpdir="`mktemp -d`"
    cat <<EOM > "$tmpdir/extractkey.java"
import java.security.Key;
import java.security.KeyStore;
import java.io.FileDescriptor;
import java.io.FileInputStream;
import java.io.FileOutputStream;

class extractkey
{
    static public void main(String[] args) throws Exception
    {
        String fname = "$1";
        String alias = "$2";
        char[] password = new String("$3").toCharArray();

        KeyStore store = KeyStore.getInstance("JKS");
        FileInputStream fin = new FileInputStream(fname);
        store.load(fin, password);
        Key key = store.getKey(alias, password);
        FileOutputStream fout = new FileOutputStream(FileDescriptor.out);
        fout.write(key.getEncoded());
    }
}
EOM
    javac "$tmpdir/extractkey.java"

    echo "-----BEGIN PRIVATE KEY-----"
    java -cp "$tmpdir" extractkey | base64
    echo "-----END PRIVATE KEY-----"

    rm -rf -- "$tmpdir"
}
"""



#
# Copyright (c) 2010 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#           http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import inspect

from ovirtsdk.xml import params
from ovirtsdk.infrastructure import brokers
from ovirtcli.utils.methodhelper import MethodHelper

class TypeHelper():
    __known_wrapper_types = None
    __known_decorators_types = None

    @staticmethod
    def _getKnownTypes():
        known_wrapper_types = {}
        for name, obj in inspect.getmembers(params):
            if inspect.isclass(obj) and (isinstance(obj, type(params.BaseResource)) or
                                         isinstance(obj, type(params.BaseResources)) or
                                         isinstance(obj, type(params.GeneratedsSuper)) or
                                         isinstance(obj, type(params.BaseDevices)) or
                                         isinstance(obj, type(params.BaseDevice))):
                known_wrapper_types[name.lower()] = name
        return known_wrapper_types

    @staticmethod
    def _getKnownDecoratorsTypes():
        __known_decorators_types = {}
        for name, obj in inspect.getmembers(brokers):
            if inspect.isclass(obj):
                __known_decorators_types[name.lower()] = name
        return __known_decorators_types

    @staticmethod
    def isKnownType(typ):
        if TypeHelper.__known_wrapper_types == None:
            TypeHelper.__known_wrapper_types = TypeHelper._getKnownTypes()
        return TypeHelper.__known_wrapper_types.has_key(typ.lower())

    @staticmethod
    def isKnownDecoratorType(typ):
        if TypeHelper.__known_decorators_types == None:
            TypeHelper.__known_decorators_types = TypeHelper._getKnownDecoratorsTypes()
        return TypeHelper.__known_decorators_types.has_key(typ.lower())

    @staticmethod
    def getKnownTypes():
        if TypeHelper.__known_wrapper_types == None:
            TypeHelper.__known_wrapper_types = TypeHelper._getKnownTypes()
        return TypeHelper.__known_wrapper_types.values()

    @staticmethod
    def getKnownDecoratorsTypes():
        if TypeHelper.__known_decorators_types == None:
            TypeHelper.__known_decorators_types = TypeHelper._getKnownDecoratorsTypes()
        return TypeHelper.__known_decorators_types.values()

    @staticmethod
    def getDecoratorType(name):
        if name and name != '':
            if TypeHelper.__known_decorators_types == None:
                TypeHelper.__known_decorators_types = TypeHelper._getKnownDecoratorsTypes()
            if TypeHelper.__known_decorators_types.has_key(name.lower()):
                return TypeHelper.__known_decorators_types[name.lower()]
        return None

    @staticmethod
    def get_actionable_types(expendNestedTypes=False, groupOptions=False):
        """INTERNAL: return a list of actionable types."""
        types = {}
        exceptions = ['delete', 'update']

        for decorator in TypeHelper.getKnownDecoratorsTypes():
                if not decorator.endswith('s'):
                    dct = getattr(brokers, decorator).__dict__
                    if dct and len(dct) > 0:
                        for method in dct:
                            if method not in exceptions and not method.startswith('_'):
                                MethodHelper.get_method_params(brokers, decorator, '__init__', types,
                                                               expendNestedTypes=expendNestedTypes,
                                                               groupOptions=groupOptions)
                                break
        return types

    @staticmethod
    def get_types_containing_method(method, expendNestedTypes=False, groupOptions=False):
        """return a list of types by method including context in which this method available."""
        types = {}

        for decorator in TypeHelper.getKnownDecoratorsTypes():
                if not decorator.endswith('s'):
                    dct = getattr(brokers, decorator).__dict__
                    if dct and len(dct) > 0 and dct.has_key(method):
                        MethodHelper.get_method_params(brokers, decorator, '__init__', types,
                                                       expendNestedTypes=expendNestedTypes,
                                                       groupOptions=groupOptions)
        return types

    @staticmethod
    def get_types_by_method(plural, method, expendNestedTypes=False, groupOptions=False):
        """INTERNAL: return a list of types that implement given method and context/s of this types."""
        sing_types = {}

        if method:
            for decorator in TypeHelper.getKnownDecoratorsTypes():
                dct = getattr(brokers, decorator).__dict__
                if dct.has_key(method):
                    if decorator.endswith('s'):
                        cls_name = TypeHelper.getDecoratorType(decorator[:len(decorator) - 1])
                        if cls_name:
                            MethodHelper.get_method_params(brokers,
                                                           cls_name,
                                                           '__init__',
                                                           sing_types,
                                                           expendNestedTypes,
                                                           groupOptions)

            if plural:
                sing_types_plural = {}
                for k in sing_types.keys():
                    sing_types_plural[TypeHelper.to_plural(k)] = sing_types[k]
                return sing_types_plural
            return sing_types

    @staticmethod
    def to_singular(string):
        if string.endswith('s'):
            return string[:len(string) - 1]
        return string

    @staticmethod
    def to_plural(string):
        if not string.endswith('s'):
            return string + 's'
        return string



#
# Copyright (c) 2013 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#           http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class DFSAState(object):
    DISCONNECTING, DISCONNECTED, CONNECTING, \
    CONNECTED, UNAUTHORIZED, EXITING, COMMUNICATION_ERROR = range(7)

    def __init__(self, Type):
        self.value = Type

    def __str__(self):
        if self.value == DFSAState.DISCONNECTED:
            return 'DISCONNECTED'
        if self.value == DFSAState.CONNECTED:
            return 'CONNECTED'
        if self.value == DFSAState.UNAUTHORIZED:
            return 'UNAUTHORIZED'
        if self.value == DFSAState.EXITING:
            return 'EXITING'
        if self.value == DFSAState.DISCONNECTING:
            return 'DISCONNECTING'
        if self.value == DFSAState.CONNECTING:
            return 'CONNECTING'
        if self.value == DFSAState.COMMUNICATION_ERROR:
            return 'COMMUNICATION_ERROR'

    def __eq__(self, y):
        return self.value == y.value



#
# Copyright (c) 2010 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#           http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import sys
import logging
import traceback

import codecs
import cStringIO
from kitchen.text.converters import getwriter
import getpass

from StringIO import StringIO
from subprocess import Popen, PIPE

from cli.error import ParseError, CommandError
from cli.object import create
from cli.settings import Settings
from cli.parser import Parser
from cli.platform import Terminal
from cli import platform
from cli.executionmode import ExecutionMode
from ovirtcli.utils.colorhelper import ColorHelper


class ExecutionContext(object):
    """A CLI execution context."""

    name = None

    OK = 0
    SYNTAX_ERROR = 1
    COMMAND_ERROR = 2
    INTERRUPTED = 3
    UNKNOWN_ERROR = 4
    COMMUNICATION_ERROR = 5
    AUTHENTICATION_ERROR = 6

    Parser = Parser
    Terminal = Terminal
    Settings = Settings

    welcome = None
    goodbye = None

    def __init__(self, cmdin=None, args=None, option_parser=None):
        """Constructor."""
        self.parameters_cash = {}
        self.cmdin = cmdin or sys.stdin
        self.args = args
        self.option_parser = option_parser
        self.commands = []
        self.status = None
        self.mode = ExecutionMode.SHELL
        self.interactive = sys.stdin.isatty() and sys.stdout.isatty() \
                    and self.cmdin is sys.stdin
        self.settings = create(self.Settings, self.name)
        self.parser = create(self.Parser)
        self.terminal = create(self.Terminal)
        self._setup_logging()
        self._load_settings()
        self.setup_commands()
        self.__setup_pager()
        self.__encoding = None

    def _setup_logging(self):
        """Configure logging."""
        logger = logging.getLogger()
        handler = logging.StreamHandler(sys.stdout)
        formatter = logging.Formatter('%(levelname)s: %(message)s')
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        logger.setLevel(logging.INFO)
        self._logger = logger

    def _collect_connection_data(self):
        try:
            if self.settings['ovirt-shell:url'] == '' and \
            not self.__is_option_specified_in_cli_args('--url')  and \
            not self.__is_option_specified_in_cli_args('-l'):
                self.settings['ovirt-shell:url'] = raw_input('URL: ')
            if self.settings['ovirt-shell:username'] == '' and \
            not self.__is_option_specified_in_cli_args('--username') and \
            not self.__is_option_specified_in_cli_args('-u'):
                self.settings['ovirt-shell:username'] = raw_input('Username: ')
            if self.settings['ovirt-shell:password'] == '':
                self.settings['ovirt-shell:password'] = getpass.getpass()
        except KeyboardInterrupt:
            sys.exit('')
        except EOFError:
            sys.exit('')
        finally:
            sys.stdin.flush()

    def __exclude_app_options(self):
        for opt in self.option_parser.app_options:
            if self.__is_option_specified_in_cli_args(opt):
                self.__option_error(opt)

    def __option_error(self, opt):
        from cli.messages import Messages
        sys.exit('\n' + Messages.Error.NO_SUCH_OPTION % opt + '\n')

    def __is_option_specified_in_cli_args(self, optionname):
        if self.args:
            if optionname in self.args: return True
            for item in self.args:
                if item.startswith(optionname + '='):
                    return True
        return False

    def __is_auto_connect(self):
        return (self.args and \
            ('-c' in self.args or '--connect' in self.args)) or \
        self.settings.get('cli:autoconnect')

    def _load_settings(self):
        """Load settings."""
        found, old_format = self.settings.load_config_file()
        if not found:
            self.settings.write_example_config_file()
        elif old_format:
            self.settings.write_config_file()
        self.__exclude_app_options()
        self.settings.add_callback('cli:debug', self._set_debug)
        self._set_debug('cli:debug', self.settings['cli:debug'])

    def _set_debug(self, key, value):
        """Enable or disable debugging (callback)."""
        if value:
            level = logging.DEBUG
        else:
            level = logging.INFO
        self._logger.setLevel(level)

    def setup_commands(self):
        """Register the default commands. Override in a subclass."""

    def add_command(self, command):
        """Add an additional command. `command' must implement the Command
        interface."""
        for i in range(len(self.commands)):
            if self.commands[i].name == command.name:
                self.commands[i] = command
                self.parameters_cash[command.name.lower()] = {}
                break
        else:
            self.parameters_cash[command.name.lower()] = {}
            self.commands.append(command)

    def execute_loop(self):
        """Run a read/parse/execute loop until EOF."""
        if self.interactive and self.welcome:
            sys.stdout.write('%s\n' % self.welcome)
        self._eof = False
        while not self._eof:
            command = self._read_command()
            if not command:
                continue
            self.execute_string(command)
        if self.interactive and self.goodbye:
            sys.stdout.write('%s\n' % self.goodbye)

    def execute_string(self, command):
        """Parse and execute a string."""
        try:
            parsed = self.parser.parse(command)
        except EOFError:
            sys.stderr.write('error: incomplete command\n')
            return
        except ParseError, e:
            self.status = self.SYNTAX_ERROR
            sys.stderr.write('error: %s\n' % str(e))
            return
        for command in parsed:
            try:
                self._execute_command(command)
            except Exception, e:
                self._handle_exception(e)
            else:
                self.status = self.OK


    def __error_to_string(self, err):
        """Converts an exception to string and normalizing it."""
        err_patterns = [
            "[ERROR]::",
            "oVirt API connection failure, "
        ]

        err_str = str(err)
        err_str = (
           "\n" if not (
                    err_str.startswith("\n")
                    or
                    err_str.startswith("\r\n")
                )
                else ""
        ) + err_str

        for err_pattern in err_patterns:
            if err_str.find(err_pattern) <> -1:
                err_str = err_str.replace(
                              err_pattern, ""
                          )

        return err_str

    def _handle_exception(self, e):
        from ovirtsdk.infrastructure.errors import \
            RequestError, ConnectionError, AmbiguousQueryError

        """Handle an exception. Can be overruled in a subclass."""
        if isinstance(e, KeyboardInterrupt):
            self.status = self.INTERRUPTED
            sys.stdout.write('\n')
        elif isinstance(e, RequestError):
            if hasattr(e, 'status') and e.status == 401:
                self.status = self.AUTHENTICATION_ERROR
            else:
                self.status = getattr(e, 'status', self.COMMAND_ERROR)
            self.__pint_error(e)
            if hasattr(e, 'help'):
                sys.stderr.write('%s\n' % e.help)
        elif isinstance(e, CommandError) or isinstance(e, AmbiguousQueryError):
            self.status = getattr(e, 'status', self.COMMAND_ERROR)
            self.__pint_error(e)
            if hasattr(e, 'help'):
                sys.stderr.write('%s\n' % e.help)
        elif isinstance(e, SyntaxError):
            self.status = getattr(e, 'status', self.SYNTAX_ERROR)
            self.__pint_error(e)
            if hasattr(e, 'help'):
                sys.stderr.write('%s\n' % e.help)
        elif isinstance(e, ConnectionError):
            self.status = getattr(e, 'status', self.COMMUNICATION_ERROR)
            self.__pint_error(e)
            if hasattr(e, 'help'):
                sys.stderr.write('%s\n' % e.help)
        elif isinstance(e, Warning):
            self.__pint_warning(e)
        else:
            self.status = self.UNKNOWN_ERROR
            if self.settings['cli:debug']:
                sys.stderr.write(traceback.format_exc())
            else:
                self.__pint_error(e, header='UNKNOWN ERROR')

    def __pint_error(self, e, header='ERROR'):
        """
        prints error to stderr

        @param e: exception
        @param header: the error header
        """

        text = self.formatter.format_terminal(
                          text=self.__error_to_string(e).strip(),
                          border='=',
                          termwidth=self.terminal._get_width(),
                          newline="\n",
                          header=header.strip()
        )

        text = ColorHelper.colorize(
                  text,
                  ColorHelper.RED if self.mode != ExecutionMode.SCRIPT
                                     and self.interactive
                                  else None
        )

        sys.stderr.write(text + "\n")

    def __pint_warning(self, e):
        """
        prints warning to stdout

        @param e: exception
        """

        text = self.formatter.format_terminal(
                          text=self.__error_to_string(e).strip(),
                          border='=',
                          termwidth=self.terminal._get_width(),
                          newline="\n",
                          header='WARNING'
        )

        text = ColorHelper.colorize(
                  text,
                  ColorHelper.YELLOW if self.mode != ExecutionMode.SCRIPT
                                        and self.interactive
                                     else None
        )

        sys.stdout.write(text + "\n")

    def _pint_text(self, text, file=sys.stdout):  # @ReservedAssignment
        """
        prints text to output device

        @param text: text to print
        @param file: the output device to fetch encoding from
        """
        encoding = self.__get_encoding(file)
        file.write(
               "\n" +
               text.encode(encoding, "replace") +
               "\n"
        )

    def __get_encoding(self, file):  # @ReservedAssignment
        """
        fetches encoding from the file

        @param file: the output device to fetch encoding from
        """
        if not self.__encoding:
            encoding = getattr(file, "encoding", None)
            if not encoding:
                encoding = sys.getdefaultencoding()
            self.__encoding = encoding
        return self.__encoding

    def _read_command(self):
        """Parse input until we can parse at least one full command, and
        return that input."""
        command = ''
        prompt = self.settings['cli:ps1']
        while True:
            try:
                if self.interactive:
                    line = self.terminal.readline(prompt)
                else:
                    line = self.cmdin.readline()
            except EOFError:
                self._eof = True
                sys.stdout.write('\n')
                return
            except KeyboardInterrupt:
                sys.stdout.write('\n')
                return
            if not line:
                self._eof = True
                sys.stdout.write('\n')
                return
            command += line
            try:
                parsed = self.parser.parse(command)  # @UnusedVariable
            except EOFError:
                prompt = self.settings['cli:ps2']
                continue
            except ParseError, e:
                self.status = self.SYNTAX_ERROR
                sys.stderr.write('error: %s\n' % str(e))
                return
            else:
                break
        return command

    def __setup_pager(self):
        pager = self.settings.get('cli:pager')
        if not pager:
            pager = platform.get_pager()
        self.__pager = pager

    def _execute_command(self, parsed):
        """INTERNAL: execute a command."""
#        if parsed[0] == '!':
#            self._execute_shell_command(parsed[1])
#            return

        # The parser produces a list of parameters that include both the
        # arguments and the options. We need to move them to separate
        # lists to ease later processing. Arguments are plain strings and
        # options are pairs containing the name of the option and the value,
        # so it is easy to separate them checking their type.
        name, parameters, redirections, pipeline = parsed
        args = []
        opts = []
        for parameter in parameters:
            if type(parameter) is str:
                args.append(parameter)
            else:
                opts.append(parameter)

        if self.settings.get('cli:autopage'):
            if pipeline:
                pipeline += '| %s' % self.__pager
            else:
                pipeline = self.__pager
        command = self._create_command(name, args, opts)
        self.command = command
        self._setup_pipeline(pipeline)
        self._setup_io_streams(redirections)
        try:
            command.run(self)
        finally:
            self._restore_io_streams()
            self._execute_pipeline()

    def _execute_shell_command(self, command):
        """INTERNAL: execute a shell command."""
        shell = Popen(command, shell=True)
        retcode = shell.wait()
        if retcode != 0:
            m = 'shell command exited with error'
            raise CommandError(m, status=retcode)

    def _create_command(self, name, args, opts):
        """INTERNAL: instantiate a new command."""
        cls = self._get_command(name)
        if cls is None:
            raise CommandError, 'unknown command: %s' % name
        return cls(args, opts)

    def _get_command(self, name):
        """Return the command class for `name' or None."""
        for cls in self.commands:
            if cls.name == name:
                return cls
            if name in cls.aliases:
                return cls

    def _setup_pipeline(self, pipeline):
        """INTERNAL: set up the pipeline, if any."""
        if not pipeline:
            self._pipeline = None
            return

        self._pipeline = Popen(pipeline, stdin=PIPE, stderr=PIPE, shell=True)
        self._pipeinput = codecs.getwriter("utf8")(cStringIO.StringIO())
        self.terminal.stdout = self._pipeinput

    def _setup_io_streams(self, redirections=[]):
        """INTERNAL: set up standard input/output/error."""
        for type_, arg in redirections:
            if type_ == '<':
                self.terminal.stdin = file(arg)
            elif type_ == '<<':
                self.terminal.stdin = StringIO(arg)
            elif type_ == '>':
                self.terminal.stdout = file(arg, 'w')
            elif type_ == '>>':
                self.terminal.stdout = file(arg, 'a')

    def _restore_io_streams(self):
        """INTERNAL: reset IO streams."""
        self.terminal.stdin = sys.stdin
        self.terminal.stdout = getwriter('utf8')(sys.stdout)
        self.terminal.stderr = sys.stderr

    def _execute_pipeline(self):
        """INTERNAL: execute the command pipleine (if any)."""
        if not self._pipeline:
            return
        pipeinput = self._pipeinput.getvalue()
        dummy, stderr = self._pipeline.communicate(pipeinput)
        retcode = self._pipeline.returncode
        if retcode != 0:
            raise CommandError, stderr.rstrip()



#
# Copyright (c) 2010 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#           http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from cli import compat
import types

from ovirtcli.annotations.requires import Requires
from ovirtcli.state.dfsastate import DFSAState

class Error(Exception):
    """Base class for python-cli errors."""

    def __init__(self, message=None, **kwargs):
        if message is None:
            message = self.__doc__
        message = message + '\n' if not message.endswith('\n') \
                                 else message
        compat.super(Error, self).__init__(message)
        for key in kwargs:
            setattr(self, key, kwargs[key])


class ParseError(Error):
    """Error parsing command line."""


class CommandError(Error):
    """Illegal command."""

class SyntaxError(Error):  # @ReservedAssignment
    """Illegal syntax."""

class StateError(Error):
    """raised when state change error occurs."""
#     @Requires([DFSAEvent, types.StringType])
    def __init__(self, destination, current):
        """
        @param destination: the destination DFSAEvent
        @param current: the current state
        """
        super(StateError, self).__init__(
             message=
             (
                '\n\nMoving to the "%s" state is not allowed,\n'
                %
                str(DFSAState(destination.get_destination()))
             )
             +
             (
                'eligible states from the "%s" state are:\n%s'
                %
                (
                    str(DFSAState(current)),
                    str([ str(DFSAState(src))
                      for src in destination.get_sources()
                      ]
                    )
                 )
             )
        )

class UnknownEventError(Error):
    """raised when unregistered event is triggered."""
#     @Requires(types.StringType)
    def __init__(self, name):
        """
        @param name: the name of DFSAEvent
        """
        super(UnknownEventError, self).__init__(
             message=(
                'Event %s, was not properly registered.' % \
                name
             )
        )



#
# Copyright (c) 2010 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#           http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import os
from cli.error import Error
from ovirtcli.platform import util
from cli.messages import Messages
from subprocess import Popen, PIPE


def launch_vnc_viewer(host, port, ticket, debug=False):
    """Launch a VNC viewer on host::port with `password'."""
    display = os.environ.get('DISPLAY')
    if display is None:
        raise Error, Messages.Error.INVALID_ENV_MODE_FOR_CONSOLE % 'vnc'
    cmd = util.which('vncviewer')
    if cmd is None:
        raise Error, Messages.Error.NO_CONSOLE_FOUND % ('vnc', 'vnc')
    cmd_passwd = util.which('vncpasswd')
    if cmd_passwd is None:
        raise Error, Messages.Error.NO_SUCH_COMMAND % 'vncpasswd'
    p = Popen([cmd_passwd, "-f"], shell=False, stdin=PIPE, stdout=PIPE)
    password = p.communicate(input=ticket)[0]
    args = [cmd, '%s::%s' % (host, port), '-passwordFile', '/dev/stdin' ]
    pid, pstdin = util.spawn(cmd, args, debug)
    os.write(pstdin, password)
    os.close(pstdin)
